
import os

import predix.config
import predix.security.uaa
import predix.security.acs
import predix.admin.service


class AccessControl(object):
    """
    Access Control provides attribute based access control.
    """
    def __init__(self, name=None, uaa=None, *args, **kwargs):
        super(AccessControl, self).__init__(*args, **kwargs)
        self.service_name = 'predix-acs'
        self.plan_name = 'Free'
        self.use_class = predix.security.acs.AccessControl

        self.service = predix.admin.service.PredixService(self.service_name,
                self.plan_name, name=name, uaa=uaa)

    def exists(self):
        """
        Returns whether or not this service already exists.
        """
        return self.service.exists()

    def create(self):
        """
        Create an instance of the Access Control Service with the typical
        starting settings.
        """
        self.service.create()

        # Set environment variables for immediate use
        uri = predix.config.get_env_key(self.use_class, 'uri')
        os.environ[uri] = self.service.settings.data['uri']

        zone_id = predix.config.get_env_key(self.use_class, 'zone_id')
        os.environ[zone_id] = self.service.settings.data['zone']['http-header-value']

    def grant_client(self, client_id):
        """
        Grant the given client id all the scopes and authorities
        needed to work with the access control service.
        """
        zone = self.service.settings.data['zone']['oauth-scope']

        scopes = ['openid', zone,
                  'acs.policies.read', 'acs.attributes.read',
                  'acs.policies.write', 'acs.attributes.write']

        authorities = ['uaa.resource', zone,
                  'acs.policies.read', 'acs.policies.write',
                  'acs.attributes.read', 'acs.attributes.write']

        self.service.uaa.uaac.update_client_grants(client_id, scope=scopes,
                authorities=authorities)

        return self.service.uaa.uaac.get_client(client_id)

    def add_to_manifest(self, manifest):
        """
        Add useful details to the manifest about this service
        so that it can be used in an application.

        :param manifest: An predix.admin.app.Manifest object
            instance that manages reading/writing manifest config
            for a cloud foundry app.
        """
        # Add this service to list of services
        manifest.add_service(self.service.name)

        # Add environment variables
        uri = predix.config.get_env_key(self.use_class, 'uri')
        manifest.add_env_var(uri,
                self.service.settings.data['uri'])

        zone_id = predix.config.get_env_key(self.use_class, 'zone_id')
        manifest.add_env_var(zone_id, 
                self.service.settings.data['zone']['http-header-value'])

        manifest.write_manifest()




import os

import predix.config
import predix.admin.service

import predix.security.uaa


class UserAccountAuthentication(object):
    """
    User Account and Authentication Serice (UAA) provides user management
    for Cloud Foundry.

    A few helpful notes:
    - You must already be logged into a CF target environment
    - There can only be 10 instances of UAA within an org

    """
    def __init__(self, plan_name=None, name=None, *args, **kwargs):
        super(UserAccountAuthentication, self).__init__(*args, **kwargs)
        self.service_name = 'predix-uaa'
        self.plan_name = plan_name or 'Free'
        self.use_class = predix.security.uaa.UserAccountAuthentication

        self.service = predix.admin.service.CloudFoundryService(self.service_name,
                self.plan_name)
        self.is_admin = False

        # If UAA already created we can authenticate immediately
        if self.exists():
            self.authenticate()

    def exists(self):
        """
        Tests whether this given service already exists.
        """
        return self.service.exists()

    def create(self, secret):
        """
        Create a new instance of the UAA service.  Requires a
        secret password for the 'admin' user account.
        """
        parameters = {"adminClientSecret": secret}
        self.service.create(parameters=parameters)

        uri = predix.config.get_env_key(self.use_class, 'uri')
        os.environ[uri] = self.service.settings.data['uri']

        # Once we create it login
        self.authenticate()

    def add_to_manifest(self, manifest):
        """
        Add useful details to the manifest about this service
        so that it can be used in an application.

        :param manifest: An predix.admin.app.Manifest object
            instance that manages reading/writing manifest config
            for a cloud foundry app.
        """
        # Add this service to list of services
        manifest.add_service(self.service.name)

        # Add environment variables
        uri = predix.config.get_env_key(self.use_class, 'uri')
        manifest.add_env_var(uri, self.service.settings.data['uri'])

        manifest.write_manifest()

    def authenticate(self):
        """
        Authenticate into the UAA instance as the admin user.
        """
        uri = predix.config.get_env_key(self.use_class, 'uri')
        os.environ[uri] = self.service.settings.data['uri']

        self.uaac = predix.security.uaa.UserAccountAuthentication()
        self.uaac.authenticate('admin', self._get_admin_secret(),
                use_cache=False)
        self.is_admin = True

    def create_client(self, client_id, client_secret):
        """
        Create a new client for use by applications.
        """
        assert self.is_admin, "Must authenticate() as admin to create client"
        return self.uaac.create_client(client_id, client_secret)

    def add_client_to_manifest(self, client_id, client_secret, manifest):
        """
        Add the client credentials to the specified manifest.
        """
        assert self.is_admin, "Must authenticate() as admin to create client"
        return self.uaac.add_client_to_manifest(client_id, client_secret,
                manifest)

    def _get_admin_secret(self):
        return self.service.settings.data['adminClientSecret']

    def _get_issuer_id(self):
        return self.service.settings.data['issuerId']




import os
import yaml
import logging


class Manifest(object):
    """
    Cloud Foundry utilizes a MANIFEST.yml file as the source
    of application configuration.  As we setup services and
    run our applications the manifest is a place to store
    important configuration details.
    """
    def __init__(self, manifest_path='manifest.yml', app_name='my-predix-app', debug=False):
        self.manifest_path = os.path.expanduser(manifest_path)
        self.app_name = app_name

        # App may have a client
        self.client_id = None
        self.client_secret = None

        if debug:
            logging.basicConfig(level=logging.DEBUG)

        # Read or Generate a manifest file
        if os.path.exists(self.manifest_path):
            manifest = self.read_manifest()
        else:
            manifest = self.create_manifest()

        # Probably always want manifest loaded into environment
        self.set_os_environ()

    def read_manifest(self):
        """
        Read an existing manifest.
        """
        with open(self.manifest_path, 'r') as input_file:
            self.manifest = yaml.safe_load(input_file)
            if 'env' not in self.manifest:
                self.manifest['env'] = {}
            if 'services' not in self.manifest:
                self.manifest['services'] = []

            input_file.close()

    def create_manifest(self):
        """
        Create a new manifest and write it to
        disk.
        """
        self.manifest = {}
        self.manifest['applications'] = [{'name': self.app_name}]
        self.manifest['services'] = []
        self.manifest['env'] = {}

        self.write_manifest()

    def write_manifest(self):
        """
        Write manifest to disk.
        """
        with open(self.manifest_path, 'w') as output_file:
            yaml.safe_dump(self.manifest, output_file,
                    default_flow_style=False, explicit_start=True)
            output_file.close()

    def add_env_var(self, key, value):
        """
        Add the given key / value as another environment
        variable.
        """
        self.manifest['env'][key] = value
        os.environ[key] = value

    def add_service(self, service_name):
        """
        Add the given service to the manifest.
        """
        if service_name not in self.manifest['services']:
            self.manifest['services'].append(service_name)

    def set_os_environ(self):
        """
        Will load any environment variables found in the
        manifest file into the current process for use
        by applications.

        When apps run in cloud foundry this would happen
        automatically.
        """
        for key in self.manifest['env'].keys():
            os.environ[key] = self.manifest['env'][key]

    def get_client_id(self):
        """
        Return the client id that should have all the
        needed scopes and authorities for the services
        in this manifest.
        """
        key = 'PREDIX_APP_CLIENT_ID'
        if key not in self.manifest['env']:
            raise ValueError("%s undefined in manifest." % key)

        self.client_id = self.manifest['env'][key]
        return self.client_id

    def get_client_secret(self):
        """
        Return the client secret that should correspond with
        the client id.
        """
        key = 'PREDIX_APP_CLIENT_SECRET'
        if key not in self.manifest['env']:
            raise ValueError("%s must be added to manifest." % key)

        self.client_secret = self.manifest['env'][key]
        return self.client_secret

    def get_timeseries(self, *args, **kwargs):
        """
        Returns an instance of the Time Series Service.
        """
        import predix.data.timeseries
        ts = predix.data.timeseries.TimeSeries(*args, **kwargs)
        return ts

    def get_asset(self):
        """
        Returns an instance of the Asset Service.
        """
        import predix.data.asset
        asset = predix.data.asset.Asset()
        return asset

    def get_uaa(self):
        """
        Returns an insstance of the UAA Service.
        """
        import predix.security.uaa
        uaa = predix.security.uaa.UserAccountAuthentication()
        return uaa

    def get_acs(self):
        """
        Returns an instance of the Asset Control Service.
        """
        import predix.security.acs
        acs = predix.security.acs.AccessControl()
        return acs

    def get_weather(self):
        """
        Returns an instance of the Weather Service.
        """
        import predix.data.weather
        weather = predix.data.weather.WeatherForecast()
        return weather

    def get_blobstore(self):
        import predix.data.blobstore
        blobstore = predix.data.blobstore.BlobStore()
        return blobstore




import predix.admin.cf.api
import predix.admin.cf.spaces


class App(object):
    """
    Operations and data for Cloud Foundry Apps.
    """
    def __init__(self, *args, **kwargs):
        super(App, self).__init__(*args, **kwargs)

        self.api = predix.admin.cf.api.API()
        self.space = predix.admin.cf.spaces.Space()

    def get_app_guid(self, app_name):
        """
        Returns the GUID for the app instance with
        the given name.
        """
        summary = self.space.get_space_summary()
        for app in summary['apps']:
            if app['name'] == app_name:
                return app['guid']

    def delete_app(self, app_name):
        """
        Delete the given app.

        Will fail intentionally if there are any service
        bindings.  You must delete those first.
        """
        if app_name not in self.space.get_apps():
            logging.warn("App not found so... succeeded?")
            return True

        guid = self.get_app_guid(app_name)
        self.api.delete("/v2/apps/%s" % (guid))




import logging

import predix.admin.uaa
import predix.admin.config
import predix.admin.cf.services


class CloudFoundryService(object):
    """
    Cloud Foundry Services
    """
    def __init__(self, service_name, plan_name, name=None, *args, **kwargs):
        super(CloudFoundryService, self).__init__(*args, **kwargs)
        self.service = predix.admin.cf.services.Service()

        # Basic details needed for creating the service
        self.plan_name = plan_name
        self.service_name = service_name
        self.name = name or self._generate_name(self.service.space.name, service_name, plan_name)

        # Cache for service configuration details
        self.config_path = self._get_config_path()
        self.settings = predix.admin.config.ServiceConfig(self.config_path)

        # Handle case of missing service keys and config
        if self.exists() and not self.settings.data:
            self.settings.save(self._get_service_config())

    def _generate_name(self, space, service_name, plan_name):
        """
        Can generate a name based on the space, service name and plan.
        """
        return str.join('-', [space, service_name, plan_name]).lower()

    def _get_config_path(self):
        """
        Return a sensible configuration path for caching config
        settings.
        """
        org = self.service.space.org.name
        space = self.service.space.name
        name = self.name

        return "~/.predix/%s/%s/%s.json" % (org, space, name)

    def _create_service(self, parameters={}):
        """
        Create a Cloud Foundry service that has custom parameters.
        """
        logging.debug("_create_service()")
        logging.debug(str.join(',', [self.service_name, self.plan_name,
            self.name, str(parameters)]))
        return self.service.create_service(self.service_name, self.plan_name,
                self.name, parameters)

    def _delete_service(self, service_only=False):
        """
        Delete a Cloud Foundry service and any associations.
        """
        logging.debug('_delete_service()')
        return self.service.delete_service(self.service_name)

    def _get_or_create_service_key(self):
        """
        Get a service key or create one if needed.
        """
        keys = self.service._get_service_keys(self.name)
        for key in keys['resources']:
            if key['entity']['name'] == self.service_name:
                return self.service.get_service_key(self.name,
                        self.service_name)

        self.service.create_service_key(self.name, self.service_name)
        return self.service.get_service_key(self.name, self.service_name)

    def _get_service_config(self):
        """
        Will get configuration for the service from a service key.
        """
        key = self._get_or_create_service_key()

        config = {}
        config['service_key'] = [{'name': self.name}]
        config.update(key['entity']['credentials'])

        return config

    def exists(self):
        """
        Test whether or not this service already exists.
        """
        return self.service.space.has_service_with_name(self.name)

    def create(self, parameters={}):
        """
        Create the service.
        """
        # Create the service
        cs = self._create_service(parameters=parameters)

        # Create the service key to get config details and
        # store in local cache file.
        cfg = parameters
        cfg.update(self._get_service_config())
        self.settings.save(cfg)


class PredixService(CloudFoundryService):
    """
    Predix Services extend Cloud Foundry Services by providing
    UAA protections in some standard ways.
    """
    def __init__(self, service_name, plan_name, name=None, uaa=None, *args, **kwargs):
        super(PredixService, self).__init__(service_name, plan_name, name=name, *args, **kwargs)

        # We will create a UAA instance if not given one and authenticate
        self.uaa = self._get_or_create_uaa(uaa)
        if not self.uaa.exists():
            raise ValueError("Instance of predix-uaa is required, create one first.")

    def _get_or_create_uaa(self, uaa):
        """
        Returns a valid UAA instance for performing administrative functions
        on services.
        """
        if isinstance(uaa, predix.admin.uaa.UserAccountAuthentication):
            return uaa

        logging.debug("Initializing a new UAA")
        return predix.admin.uaa.UserAccountAuthentication()

    def create(self, parameters={}):
        """
        Create an instance of the US Weather Forecast Service with
        typical starting settings.
        """
        # Add parameter during create for UAA issuer
        uri = self.uaa.service.settings.data['uri'] + '/oauth/token'
        parameters["trustedIssuerIds"] = [uri]
        super(PredixService, self).create(parameters=parameters)




import logging

import predix.admin.cf.api
import predix.admin.cf.spaces


class Service(object):
    """
    Operations and data for working with Cloud Foundry services.
    """
    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)

        self.api = predix.admin.cf.api.API()
        self.space = predix.admin.cf.spaces.Space()

    def get_services(self):
        """
        Get the marketplace services.
        """
        return self.api.get('/v2/services')

    def get_instance_guid(self, service_name):
        """
        Returns the GUID for the service instance with
        the given name.
        """
        summary = self.space.get_space_summary()
        for service in summary['services']:
            if service['name'] == service_name:
                return service['guid']

        raise ValueError("No service with name '%s' found." % (service_name))

    def _get_service_bindings(self, service_name):
        """
        Return the service bindings for the service instance.
        """
        instance = self.get_instance(service_name)
        return self.api.get(instance['service_bindings_url'])

    def delete_service_bindings(self, service_name):
        """
        Remove service bindings to applications.
        """
        instance = self.get_instance(service_name)
        return self.api.delete(instance['service_bindings_url'])

    def _get_service_keys(self, service_name):
        """
        Return the service keys for the given service.
        """
        guid = self.get_instance_guid(service_name)
        uri = "/v2/service_instances/%s/service_keys" % (guid)
        return self.api.get(uri)

    def get_service_keys(self, service_name):
        """
        Returns a flat list of the names of the service keys
        for the given service.
        """
        keys = []
        for key in self._get_service_keys(service_name)['resources']:
            keys.append(key['entity']['name'])

        return keys

    def has_key(self, service_name, key_name):
        """
        Tests if the given service has a key of the given name.
        """
        return key_name in self.get_service_keys(service_name)

    def get_service_key(self, service_name, key_name):
        """
        Returns the service key details.

        Similar to `cf service-key`.
        """
        for key in self._get_service_keys(service_name)['resources']:
            if key_name == key['entity']['name']:
                guid = key['metadata']['guid']

                uri = "/v2/service_keys/%s" % (guid)
                return self.api.get(uri)

        return None

    def create_service_key(self, service_name, key_name):
        """
        Create a service key for the given service.
        """
        if self.has_key(service_name, key_name):
            logging.warn("Reusing existing service key %s" % (key_name))
            return self.get_service_key(service_name, key_name)

        body = {
            'service_instance_guid': self.get_instance_guid(service_name),
            'name': key_name
            }

        return self.api.post('/v2/service_keys', body)

    def delete_service_key(self, service_name, key_name):
        """
        Delete a service key for the given service.
        """
        key = self.get_service_key(service_name, key_name)
        logging.info("Deleting service key %s for service %s" % (key, service_name))
        return self.api.delete(key['metadata']['url'])

    def get_instance(self, service_name):
        """
        Retrieves a service instance with the given name.
        """
        for resource in self.space._get_instances()['resources']:
            if resource['entity']['name'] == service_name:
                return resource['entity']

    def get_service_plans(self):
        """
        Get the available service plans.
        """
        return self.api.get('/v2/service_plans')

    def get_service_plan_for_service(self, service_name):
        """
        Return the service plans available for a given service.
        """
        services = self.get_services()
        for service in services['resources']:
            if service['entity']['label'] == service_name:
                response = self.api.get(service['entity']['service_plans_url'])
                return response['resources']

    def get_service_plan_guid(self, service_name, plan_name):
        """
        Return the service plan GUID for the given service / plan.
        """
        for plan in self.get_service_plan_for_service(service_name):
            if plan['entity']['name'] == plan_name:
                return plan['metadata']['guid']

        return None

    def create_service(self, service_type, plan_name, service_name, params):
        """
        Create a service instance.
        """
        if self.space.has_service_with_name(service_name):
            logging.warn("Service already exists with that name.")
            return self.get_instance(service_name)

        if self.space.has_service_of_type(service_type):
            logging.warn("Service type already exists.")

        guid = self.get_service_plan_guid(service_type, plan_name)
        if not guid:
            raise ValueError("No service plan named: %s" % (plan_name))

        body = {
            'name': service_name,
            'space_guid': self.space.guid,
            'service_plan_guid': guid,
            'parameters': params
            }

        return self.api.post('/v2/service_instances?accepts_incomplete=false',
                body)

    def delete_service(self, service_name, params=None):
        """
        Delete the service of the given name.  It may fail if there are
        any service keys or app bindings.  Use obliterate() if you want
        to delete it all.
        """
        if not self.space.has_service_with_name(service_name):
            logging.warn("Service not found so... succeeded?")
            return True

        guid = self.get_instance_guid(service_name)
        logging.info("Deleting service %s with guid %s" % (service_name, guid))

        # MAINT: this endpoint changes in newer version of api
        return self.api.delete("/v2/service_instances/%s?accepts_incomplete=true" %
            (guid), params=params)

    def purge(self, service_name):
        """
        Remove the service and anything that prevents from its removal such
        as service keys and app bindings.
        """
        return self.delete_service(service_name, params={'recursive': 'true'})




import json
import logging
import requests

import predix.admin.cf.config


class API(object):
    """
    Manages Cloud Foundry API calls for system administration.

    Utilizes the local cf cli to cache identity and target details.  As
    a result developers must first `cf login`.

    More details:
    http://apidocs.cloudfoundry.org/
    """

    US_WEST = 'https://api.system.aws-usw02-pr.ice.predix.io'

    def __init__(self, *args, **kwargs):
        super(API, self).__init__(*args, **kwargs)

        self.config = predix.admin.cf.config.Config()
        self.session = requests.Session()

    def _get_headers(self, content_type='application/json'):
        """
        Returns headers needed for talking to the Cloud Foundry
        Provider API endpoint.
        """
        headers = {
            'Accept': 'application/json',
            'Content-Type': content_type,
            'Authorization': self.config.get_access_token()
            }
        return headers

    def _post_headers(self):
        """
        Returns headers needed for talking to the Cloud Foundry
        Provider API endpoing for POST requests.
        """
        return self._get_headers(content_type='application/x-www-form-urlencoded')

    def get(self, path):
        """
        Generic GET with headers
        """
        uri = self.config.get_target() + path
        headers = self._get_headers()

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))

        response = self.session.get(uri, headers=headers)
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 401:
            raise predix.admin.cf.config.CloudFoundryLoginError('token invalid')
        else:
            response.raise_for_status()

    def post(self, path, data):
        """
        Generic POST with headers
        """
        uri = self.config.get_target() + path
        headers = self._post_headers()

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))

        response = self.session.post(uri, headers=headers,
                data=json.dumps(data))
        if response.status_code in (200, 201):
            return response.json()
        elif response.status_code == 401:
            raise predix.admin.cf.config.CloudFoundryLoginError('token invalid')
        else:
            logging.debug("STATUS=" + str(response.status_code))
            logging.debug("CONTENT=" + str(response.content))
            response.raise_for_status()

    def put(self, path, data):
        """
        Generic PUT with headers
        """
        uri = self.config.get_target() + path
        headers = self._get_headers()

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))

        response = self.session.put(uri, headers=headers,
                data=json.dumps(data))

        if response.status_code in (200, 201):
            return response
        elif response.status_code == 401:
            raise predix.admin.cf.config.CloudFoundryLoginError('token invalid')
        else:
            logging.debug("STATUS=" + str(response.status_code))
            logging.debug("CONTENT=" + str(response.content))
            response.raise_for_status()    

    def delete(self, path, data=None, params=None):
        """
        Generic DELETE with headers
        """
        uri = self.config.get_target() + path
        headers = {
            'Authorization': self.config.get_access_token()
            }

        logging.debug("URI=" + str(uri))
        logging.debug("HEADERS=" + str(headers))

        response = self.session.delete(
            uri, headers=headers, params=params, data=json.dumps(data))

        if response.status_code == 204:
            return response
        else:
            logging.debug("STATUS=" + str(response.status_code))
            logging.debug("CONTENT=" + str(response.content))
            response.raise_for_status()

    def get_info(self):
        """
        Return API info such as:

            - api_version
            - authorization_endpoint
            - doppler_logging_endpoint
            - logging_endpoint
            - min_cli_version
        """
        return self.get('/v2/info')




def get_env_key(obj, key=None):
    """
    Return environment variable key to use for lookups within a
    namespace represented by the package name.
    """
    return str.join('_', [obj.__module__.replace('.','_').upper(),
        key.upper()])
